package dev.fyloz.colorrecipesexplorer.service.files

import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import dev.fyloz.colorrecipesexplorer.exception.RestException
import dev.fyloz.colorrecipesexplorer.utils.File
import dev.fyloz.colorrecipesexplorer.utils.FilePath
import dev.fyloz.colorrecipesexplorer.utils.withFileAt
import mu.KotlinLogging
import org.slf4j.Logger
import org.springframework.core.io.ByteArrayResource
import org.springframework.core.io.Resource
import org.springframework.http.HttpStatus
import org.springframework.stereotype.Service
import org.springframework.web.multipart.MultipartFile
import java.io.IOException

/** Banned path shards. These are banned because they can allow access to files outside the data directory. */
val BANNED_FILE_PATH_SHARDS = setOf(
    "~",
    "..",
    "//"
)

interface FileService {
    /** Checks if the file at the given [path] exists. */
    fun exists(path: String): Boolean

    /** Reads the file at the given [path]. */
    fun read(path: String): Resource

    /** Completes the path of the given [String] by adding the working directory. */
    fun String.fullPath(): FilePath
}

interface WriteableFileService : FileService {
    /** Creates a file at the given [path]. */
    fun create(path: String)

    /** Writes the given [file] to the given [path]. If the file already exists, it will be overwritten if [overwrite] is enabled. */
    fun write(file: MultipartFile, path: String, overwrite: Boolean)

    /** Writes the given [data] to the given [path]. If the file at the path already exists, it will be overwritten if [overwrite] is enabled. */
    fun write(data: ByteArrayResource, path: String, overwrite: Boolean)

    /** Deletes the file at the given [path]. */
    fun delete(path: String)
}

@Service
class FileServiceImpl(
    private val creProperties: CreProperties
) : WriteableFileService {
    private val logger = KotlinLogging.logger {}

    override fun exists(path: String): Boolean {
        val fullPath = path.fullPath()
        return if (fullPath in FileExistCache) {
            FileExistCache.exists(fullPath)
        } else {
            withFileAt(fullPath) {
                (this.exists() && this.isFile).also {
                    FileExistCache.set(fullPath, it)
                }
            }
        }
    }

    override fun read(path: String) = ByteArrayResource(
        withFileAt(path.fullPath()) {
            if (!exists(path)) throw FileNotFoundException(path)
            try {
                readBytes()
            } catch (ex: IOException) {
                FileReadException(path).logAndThrow(ex, logger)
            }
        }
    )

    override fun create(path: String) {
        val fullPath = path.fullPath()
        if (!exists(path)) {
            try {
                withFileAt(fullPath) {
                    this.create()
                    FileExistCache.setExists(fullPath)

                    logger.info("Created file at '${fullPath.path}'")
                }
            } catch (ex: IOException) {
                FileCreateException(path).logAndThrow(ex, logger)
            }
        }
    }

    override fun write(file: MultipartFile, path: String, overwrite: Boolean) =
        prepareWrite(path, overwrite) {
            logWrittenDataSize(file.size)
            file.transferTo(this.toPath())
        }

    override fun write(data: ByteArrayResource, path: String, overwrite: Boolean) =
        prepareWrite(path, overwrite) {
            logWrittenDataSize(data.contentLength())
            this.writeBytes(data.byteArray)
        }

    override fun delete(path: String) {
        try {
            val fullPath = path.fullPath()
            withFileAt(fullPath) {
                if (!exists(path)) throw FileNotFoundException(path)

                this.delete()
                FileExistCache.setDoesNotExists(fullPath)

                logger.info("Deleted file at '${fullPath.path}'")
            }
        } catch (ex: IOException) {
            FileDeleteException(path).logAndThrow(ex, logger)
        }
    }

    override fun String.fullPath(): FilePath {
        BANNED_FILE_PATH_SHARDS
            .firstOrNull { this.contains(it) }
            ?.let { throw InvalidFilePathException(this, it) }

        return FilePath("${creProperties.dataDirectory}/$this")
    }

    private fun prepareWrite(path: String, overwrite: Boolean, op: File.() -> Unit) {
        val fullPath = path.fullPath()

        if (exists(path)) {
            if (!overwrite) throw FileExistsException(path)
        } else {
            create(path)
        }

        try {
            withFileAt(fullPath) {
                this.op()

                logger.info("Wrote data to file at '${fullPath.path}'")
            }
        } catch (ex: IOException) {
            FileWriteException(path).logAndThrow(ex, logger)
        }
    }

    private fun logWrittenDataSize(size: Long) {
        logger.debug("Writing $size bytes to file system...")
    }
}

private const val FILE_IO_EXCEPTION_TITLE = "File IO error"

class InvalidFilePathException(val path: String, val fragment: String) :
    RestException(
        "invalid-filepath",
        "Invalid file path",
        HttpStatus.BAD_REQUEST,
        "The given path is invalid because it contains a potentially malicious String '$fragment'",
        mapOf(
            "path" to path,
            "invalidString" to fragment
        )
    )

class FileExistsException(val path: String) :
    RestException(
        "io-exists",
        FILE_IO_EXCEPTION_TITLE,
        HttpStatus.BAD_REQUEST,
        "Could not write file to '$path' because it already exists. To overwrite the file set the overwrite parameter to true",
        pathMap(path)
    )

class FileNotFoundException(val path: String) :
    RestException(
        "io-notfound",
        FILE_IO_EXCEPTION_TITLE,
        HttpStatus.NOT_FOUND,
        "Could not access file at '$path' because it does not exists",
        pathMap(path)
    )

sealed class FileIOException(type: String, details: String, val path: String) :
    RestException(
        "io-$type",
        FILE_IO_EXCEPTION_TITLE,
        HttpStatus.INTERNAL_SERVER_ERROR,
        details,
        pathMap(path)
    )

class FileReadException(path: String) :
    FileIOException(
        "read",
        "Could not read file at '$path'",
        path
    )

class FileWriteException(path: String) :
    FileIOException(
        "write",
        "Could not write file to '$path'",
        path
    )

class FileCreateException(path: String) :
    FileIOException(
        "create",
        "Could not create file at '$path'",
        path
    )

class FileDeleteException(path: String) :
    FileIOException(
        "delete",
        "Could not delete file at '$path'",
        path
    )

private fun pathMap(path: String) =
    mapOf("path" to path)

private fun <T : FileIOException> T.logAndThrow(baseException: Exception, logger: Logger): Nothing {
    logger.error(this.details, baseException)
    throw this
}
