package dev.fyloz.colorrecipesexplorer.service.files

import dev.fyloz.colorrecipesexplorer.config.properties.CreProperties
import dev.fyloz.colorrecipesexplorer.utils.File
import io.mockk.*
import org.junit.jupiter.api.AfterEach
import org.junit.jupiter.api.BeforeEach
import org.junit.jupiter.api.Test
import org.junit.jupiter.api.assertThrows
import org.springframework.mock.web.MockMultipartFile
import java.io.IOException
import java.nio.file.Path
import kotlin.test.assertEquals
import kotlin.test.assertFalse
import kotlin.test.assertTrue

private val creProperties = CreProperties().apply {
    dataDirectory = "data"
}
private const val mockFilePath = "existingFile"
private val mockFilePathPath = Path.of(mockFilePath)
private val mockFileData = byteArrayOf(0x1, 0x8, 0xa, 0xf)

class FileServiceTest {
    private val fileService = spyk(FileServiceImpl(creProperties))
    private val mockFile = mockk<File> {
        every { file } returns mockk()
        every { exists() } returns true
        every { isFile } returns true
        every { toPath() } returns mockFilePathPath
    }
    private val mockMultipartFile = spyk(MockMultipartFile(mockFilePath, mockFileData))

    @BeforeEach
    internal fun beforeEach() {
        mockkObject(File.Companion)
        every { File.from(any<String>()) } returns mockFile
    }

    @AfterEach
    internal fun afterEach() {
        clearAllMocks()
    }

    private fun whenFileCached(cached: Boolean = true, test: () -> Unit) {
        mockkObject(FileExistCache) {
            every { FileExistCache.contains(any()) } returns cached

            test()
        }
    }

    private fun whenFileNotCached(test: () -> Unit) {
        whenFileCached(false, test)
    }

    private fun whenMockFilePathExists(exists: Boolean = true, test: () -> Unit) {
        every { fileService.exists(mockFilePath) } returns exists
        test()
    }

    // exists()

    @Test
    fun `exists() returns true when the file at the given path exists and is a file`() {
        whenFileNotCached {
            assertTrue { fileService.exists(mockFilePath) }

            verify {
                mockFile.exists()
                mockFile.isFile
            }
            confirmVerified(mockFile)
        }
    }

    @Test
    fun `exists() returns false when the file at the given path does not exist`() {
        whenFileNotCached {
            every { mockFile.exists() } returns false

            assertFalse { fileService.exists(mockFilePath) }

            verify {
                mockFile.exists()
            }
            confirmVerified(mockFile)
        }
    }

    @Test
    fun `exists() returns false when the file at the given path is not a file`() {
        whenFileNotCached {
            every { mockFile.isFile } returns false

            assertFalse { fileService.exists(mockFilePath) }

            verify {
                mockFile.exists()
                mockFile.isFile
            }
            confirmVerified(mockFile)
        }
    }

    @Test
    fun `exists() returns true when the file at the given path is cached as existing`() {
        whenFileCached {
            every { FileExistCache.exists(any()) } returns true

            assertTrue { fileService.exists(mockFilePath) }

            verify {
                FileExistCache.contains(any())
                FileExistCache.exists(any())

                mockFile wasNot called
            }
            confirmVerified(FileExistCache, mockFile)
        }
    }

    @Test
    fun `exists() returns false when the file at the given path is cached as not existing`() {
        whenFileCached {
            every { FileExistCache.exists(any()) } returns false

            assertFalse { fileService.exists(mockFilePath) }

            verify {
                FileExistCache.contains(any())
                FileExistCache.exists(any())

                mockFile wasNot called
            }
            confirmVerified(FileExistCache, mockFile)
        }
    }

    // read()

    @Test
    fun `read() returns a valid ByteArrayResource`() {
        whenMockFilePathExists {
            mockkStatic(File::readBytes)
            every { mockFile.readBytes() } returns mockFileData

            val redResource = fileService.read(mockFilePath)

            assertEquals(mockFileData, redResource.byteArray)
        }
    }

    @Test
    fun `read() throws FileNotFoundException when no file exists at the given path`() {
        whenMockFilePathExists(false) {
            with(assertThrows<FileNotFoundException> { fileService.read(mockFilePath) }) {
                assertEquals(mockFilePath, this.path)
            }
        }
    }

    @Test
    fun `read() throws FileReadException when an IOException is thrown`() {
        whenMockFilePathExists {
            mockkStatic(File::readBytes)
            every { mockFile.readBytes() } throws IOException()

            with(assertThrows<FileReadException> { fileService.read(mockFilePath) }) {
                assertEquals(mockFilePath, this.path)
            }
        }
    }

    // create()

    @Test
    fun `create() creates a file at the given path`() {
        whenMockFilePathExists(false) {
            whenFileNotCached {
                mockkStatic(File::create) {
                    every { mockFile.create() } just Runs
                    every { FileExistCache.setExists(any()) } just Runs

                    fileService.create(mockFilePath)

                    verify {
                        mockFile.create()

                        FileExistCache.setExists(any())
                    }
                    confirmVerified(mockFile, FileExistCache)
                }
            }
        }
    }

    @Test
    fun `create() does nothing when a file already exists at the given path`() {
        whenMockFilePathExists {
            fileService.create(mockFilePath)

            verify(exactly = 0) {
                mockFile.create()
            }
        }
    }

    @Test
    fun `create() throws FileCreateException when the file creation throws an IOException`() {
        whenMockFilePathExists(false) {
            mockkStatic(File::create)
            every { mockFile.create() } throws IOException()

            with(assertThrows<FileCreateException> { fileService.create(mockFilePath) }) {
                assertEquals(mockFilePath, this.path)
            }
        }
    }

    // write()

    @Test
    fun `write() creates and writes the given MultipartFile to the file at the given path`() {
        whenMockFilePathExists(false) {
            every { fileService.create(mockFilePath) } just Runs
            every { mockMultipartFile.transferTo(mockFilePathPath) } just Runs

            fileService.write(mockMultipartFile, mockFilePath, false)

            verify {
                fileService.create(mockFilePath)
                mockMultipartFile.transferTo(mockFilePathPath)
            }
        }
    }

    @Test
    fun `write() throws FileExistsException when a file at the given path already exists and overwrite is disabled`() {
        whenMockFilePathExists {
            with(assertThrows<FileExistsException> { fileService.write(mockMultipartFile, mockFilePath, false) }) {
                assertEquals(mockFilePath, this.path)
            }
        }
    }

    @Test
    fun `write() writes the given MultipartFile to an existing file when overwrite is enabled`() {
        whenMockFilePathExists {
            every { mockMultipartFile.transferTo(mockFilePathPath) } just Runs

            fileService.write(mockMultipartFile, mockFilePath, true)

            verify {
                mockMultipartFile.transferTo(mockFilePathPath)
            }
        }
    }

    @Test
    fun `write() throws FileWriteException when writing the given file throws an IOException`() {
        whenMockFilePathExists(false) {
            every { fileService.create(mockFilePath) } just Runs
            every { mockMultipartFile.transferTo(mockFilePathPath) } throws IOException()

            with(assertThrows<FileWriteException> {
                fileService.write(mockMultipartFile, mockFilePath, false)
            }) {
                assertEquals(mockFilePath, this.path)
            }
        }
    }

    // delete()

    @Test
    fun `delete() deletes the file at the given path`() {
        whenMockFilePathExists {
            whenFileCached {
                every { mockFile.delete() } returns true
                every { FileExistCache.setDoesNotExists(any()) } just Runs

                fileService.delete(mockFilePath)

                verify {
                    mockFile.delete()

                    FileExistCache.setDoesNotExists(any())
                }
                confirmVerified(mockFile, FileExistCache)
            }
        }
    }

    @Test
    fun `delete() throws FileNotFoundException when no file exists at the given path`() {
        whenMockFilePathExists(false) {
            with(assertThrows<FileNotFoundException> { fileService.delete(mockFilePath) }) {
                assertEquals(mockFilePath, this.path)
            }
        }
    }

    @Test
    fun `delete() throws FileDeleteException when deleting throw and IOException`() {
        whenMockFilePathExists {
            every { mockFile.delete() } throws IOException()

            with(assertThrows<FileDeleteException> { fileService.delete(mockFilePath) }) {
                assertEquals(mockFilePath, this.path)
            }
        }
    }

    // String.fullPath()

    @Test
    fun `fullPath() appends the given path to the given working directory`() {
        with(fileService) {
            val fullFilePath = mockFilePath.fullPath()

            assertEquals("${creProperties.dataDirectory}/$mockFilePath", fullFilePath.path)
        }
    }

    @Test
    fun `fullPath() throws InvalidFilePathException when the given path contains invalid fragments`() {
        with(fileService) {
            BANNED_FILE_PATH_SHARDS.forEach {
                val maliciousPath = "$it/$mockFilePath"

                with(assertThrows<InvalidFilePathException> { maliciousPath.fullPath() }) {
                    assertEquals(maliciousPath, this.path)
                    assertEquals(it, this.fragment)
                }
            }
        }
    }
}
